/*
 * JConsole.java - A java console to replace the command line window
 *                 Redirects the stdout and stderr, customizable,
 *                 can save the messages it a text file. In addition it can
 *                 auto dump and clear the textarea.
 *                 If used in conjunction with any exe creator for java,
 *                 e.g: the really great JSmooth 
 *                      http://jsmooth.sourceforge.net/
 *                 eliminates the dos application windows and your application
 *                 looks more professional.
 *
 * Copyright (c) 2004 Gregory Kotsaftis
 * gregkotsaftis@yahoo.com
 * http://zeus-jscl.sourceforge.net/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package gr.zeus.ui;

import gr.zeus.util.CurrentDateHelper;
import gr.zeus.util.IOHelper;
import java.awt.Color;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FilterOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintStream;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JOptionPane;
import javax.swing.SwingUtilities;


public class JConsole extends JFrame {
    
    // singleton implementation
    private static JConsole s_console = null;
    
    private static final PrintStream STDOUT = System.out;   // default streams
    private static final PrintStream STDERR = System.err;   // ...
    private static final Color FG_COLOR = Color.WHITE;      // foreground color
    private static final Color BG_COLOR = Color.BLACK;      // background color
    private static final Color SLT_COLOR = BG_COLOR;        // selected text color
    private static final Color SL_COLOR = FG_COLOR;         // selection color
    
    private PrintStream  m_stdoutPS = null;
    private PrintStream  m_stderrPS = null;
    private JFileChooser m_fileChooser = null;
    private String       m_confimTitle = "Question";
    private String       m_confimMessage = "The file already exists, overwrite?";
    private String       m_messagesFilename = "messages.txt"; // used by the filechooser
    private String       m_traceFilename = "trace.log";       // used by the auto dump functions
    private boolean      m_appendFirstTime = false; // overwrite trace file on first time
    private boolean      m_autoSave = false;        // don't auto save trace file
    private int          m_maxChars = -1;           // infinite characters, no limit
                                                    // NOTE: will slow down your application
                                                    //       if a lot of messages are to be
                                                    //       displayed (more than a couple of Kbytes)
    private int          m_numFlushes = 0;  // how many times we have flushed
                                            // the messages to the trace file
    
    
    
    
    /**
     * PRIVATE CONSTRUCTOR, 
     * init gui
     * prepare new streams
     */
    private JConsole()
    {
        initComponents();
        
        ta_messages.setForeground( FG_COLOR );
        ta_messages.setBackground( BG_COLOR );
        ta_messages.setSelectedTextColor( SLT_COLOR );
        ta_messages.setSelectionColor( SL_COLOR );
        
        m_stdoutPS = new PrintStream( 
                        new JTextAreaOutStream( new ByteArrayOutputStream() ) );
        m_stderrPS = new PrintStream( 
                        new JTextAreaOutStream( new ByteArrayOutputStream() ) );
        
        m_fileChooser = new JFileChooser();
        m_fileChooser.setDialogType( JFileChooser.SAVE_DIALOG );
        m_fileChooser.setFileSelectionMode( JFileChooser.FILES_ONLY );
        m_fileChooser.setMultiSelectionEnabled( false );
        m_fileChooser.setAcceptAllFileFilterUsed( true );
        m_fileChooser.setControlButtonsAreShown( true );
        
        hideConsole();
    }
    
    
    //--------------------------------------------------------------------------
    //
    //                      P U B L I C   M E T H O D S
    //
    //--------------------------------------------------------------------------
    
    /**
     * singleton
     */
    public static JConsole getConsole()
    {
        if( s_console==null )
        {
            s_console = new JConsole(); 
        }
        
        return( s_console );
    }

    
    /**
     * attach the new streams to stdout and stderr
     */
    public void startConsole()
    {        
        System.setOut( m_stdoutPS );
        System.setErr( m_stderrPS );
    }
    
    
    /**
     * attach the original streams to stdout and stderr
     */
    public void stopConsole()
    {
        System.setOut( STDOUT );
        System.setErr( STDERR );
    }
    

    /**
     * show the console
     */
    public void showConsole()
    {
        setVisible( true );        
    }
    

    /**
     * hide the console
     */
    public void hideConsole()
    {
        setVisible( false );        
    }
    

    /**
     * destruction for the console
     */
    public void destroyConsole()
    {
        hideConsole();
        
        // if we must, flush the messages to a file (force=true)
        do_it( true );
        
        // Clear the messages from JConsole (if not done already), because if we
        // have invoked destroyConsole() but not called System.exit(), if we
        // getConsole() again and execute showConsole() we will see our old
        // messages still there...
        do_btn_clear();
        
        stopConsole();
        dispose();
    }
    
    
    /**
     * single point for configuration of the JConsole
     * if any parameter is null it is not set (defaults apply)
     *
     * appendFirstTime:  if the 'traceFilename' exists the first time we try to
     *                   save the messages, should we append or overwrite?
     * autoSave:         auto saves the console messages to file 'traceFilename' when 
     *                   'maxChars' are reached and clears the messages (appends
     *                   new messages when required).
     * maxChars:         if -1 no limit, else the messages will be flushed to 
     *                   'traceFilename' and cleared when this limit is reached.
     * traceFilename:    the file used by dumpConsole() and all auto save methods.
     * messagesFilename: default filename for filechooser
     * titleText:        title for JConsole's frame 
     * fg:               foreground color for textarea
     * bg:               background color for textarea
     * slt:              selected text color for textarea
     * sl:               selection color for textarea
     * clearButtonText:  text for clear button
     * saveButtonText:   text for save button
     * closeButtonText:  text for close button
     * fcTitle:          filechooser title
     * fcButtonSaveText: filechooser approve button text
     * confimTitle:      popup window confirm overwrite file title
     * confimMessage:    popup window confirm overwrite file text
     */
    public void configureConsole(Boolean appendFirstTime,
                                 Boolean autoSave,
                                 Integer maxChars,
                                 String traceFilename,
                                 String messagesFilename,                                 
                                 String titleText,
                                 Color fg,
                                 Color bg,
                                 Color slt,
                                 Color sl,
                                 String clearButtonText,
                                 String saveButtonText,
                                 String closeButtonText,
                                 String fcTitle,
                                 String fcButtonSaveText,
                                 String confimTitle,
                                 String confimMessage)
    {
        if( appendFirstTime!=null )
        {
            m_appendFirstTime = appendFirstTime.booleanValue();
        }
        
        if( autoSave!=null )
        {
            m_autoSave = autoSave.booleanValue();
        }
        
        if( maxChars!=null )
        {
            m_maxChars = maxChars.intValue();
        }
        
        if( traceFilename!=null )
        {
            m_traceFilename = traceFilename;
        }
        
        if( messagesFilename!=null )
        {
            m_messagesFilename = messagesFilename;
        }
        
        if( titleText!=null )
        {
            setTitle( titleText );
        }
        
        if( fg!=null )
        {
            ta_messages.setForeground( fg );
        }
        
        if( bg!=null )
        {
            ta_messages.setBackground( bg );
        }
        
        if( slt!=null )
        {
            ta_messages.setSelectedTextColor( slt );
        }
        
        if( sl!=null )
        {
            ta_messages.setSelectionColor( sl );
        }
        
        if( clearButtonText!=null )
        {
            btn_clear.setText( clearButtonText );
        }
        
        if( saveButtonText!=null )
        {
            btn_save.setText( saveButtonText );
        }
        
        if( closeButtonText!=null )
        {
            btn_close.setText( closeButtonText );
        }
        
        if( fcTitle!=null )
        {
            m_fileChooser.setDialogTitle( fcTitle );
        }
        
        if( fcButtonSaveText!=null )
        {
            m_fileChooser.setApproveButtonText( fcButtonSaveText );
        }
        
        if( confimTitle!=null )
        {
            m_confimTitle = confimTitle;
        }
        
        if( confimMessage!=null )
        {
            m_confimMessage = confimMessage;
        }
    }
    
    
    /**
     * Dumps the console messages to a file.
     * Should be used for cases our program needs to exit with error and we don't
     * have time to check the console, because it is shutting down to.
     * If append is true text is appended in the file, else the file gets
     * overwritten.
     */
    public void dumpConsole(boolean append)
    {
        String timestamp = "Console trace file generated on: " +
                           String.valueOf( CurrentDateHelper.getCurrentDay() ) + "/" +
                           String.valueOf( CurrentDateHelper.getCurrentMonth() ) + "/" +
                           String.valueOf( CurrentDateHelper.getCurrentYear() ) + " " +
                           String.valueOf( CurrentDateHelper.getCurrentHour() ) + ":" +
                           String.valueOf( CurrentDateHelper.getCurrentMinutes() ) + ":" +
                           String.valueOf( CurrentDateHelper.getCurrentSeconds() + "\r\n\r\n");
        
        String completeText = null;
        if( append )
        {
            completeText = ta_messages.getText();
        }
        else
        {
            completeText = timestamp + ta_messages.getText();
        }

        try
        {
            IOHelper.saveTxtFile(m_traceFilename, completeText, append);
        }
        catch(Exception e)
        {
            // not much to do if we have error here...
            e.printStackTrace();            
        }
    }


    /**
     * Update the look and feel of the JConsole and it's components.
     * MUST be called AFTER invoking: UIManager.setLookAndFeel() in your
     * application.
     */
    public void updateLNF()
    {
        SwingUtilities.updateComponentTreeUI( this );
        SwingUtilities.updateComponentTreeUI( m_fileChooser );
        pack();
    }
    
    
    /** This method is called from within the constructor to
     * initialize the form.
     * WARNING: Do NOT modify this code. The content of this method is
     * always regenerated by the Form Editor.
     */
    private void initComponents() {//GEN-BEGIN:initComponents
        button_panel = new javax.swing.JPanel();
        btn_clear = new javax.swing.JButton();
        btn_save = new javax.swing.JButton();
        btn_close = new javax.swing.JButton();
        text_scrollpane = new javax.swing.JScrollPane();
        ta_messages = new javax.swing.JTextArea();

        setTitle("JConsole");
        button_panel.setLayout(new java.awt.FlowLayout(java.awt.FlowLayout.RIGHT));

        btn_clear.setText("Clear");
        btn_clear.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                btn_clearActionPerformed(evt);
            }
        });

        button_panel.add(btn_clear);

        btn_save.setText("Save");
        btn_save.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                btn_saveActionPerformed(evt);
            }
        });

        button_panel.add(btn_save);

        btn_close.setText("Close");
        btn_close.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                btn_closeActionPerformed(evt);
            }
        });

        button_panel.add(btn_close);

        getContentPane().add(button_panel, java.awt.BorderLayout.SOUTH);

        text_scrollpane.setBorder(new javax.swing.border.EmptyBorder(new java.awt.Insets(5, 5, 5, 5)));
        ta_messages.setColumns(80);
        ta_messages.setEditable(false);
        ta_messages.setRows(24);
        text_scrollpane.setViewportView(ta_messages);

        getContentPane().add(text_scrollpane, java.awt.BorderLayout.CENTER);

        pack();
    }//GEN-END:initComponents

    private void btn_closeActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_btn_closeActionPerformed
        
        do_btn_close();
        
    }//GEN-LAST:event_btn_closeActionPerformed

    private void btn_saveActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_btn_saveActionPerformed

        do_btn_save();
        
    }//GEN-LAST:event_btn_saveActionPerformed

    private void btn_clearActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_btn_clearActionPerformed
        
        do_btn_clear();
        
    }//GEN-LAST:event_btn_clearActionPerformed


    //--------------------------------------------------------------------------
    //
    //                      P R I V A T E   M E T H O D S
    //
    //--------------------------------------------------------------------------
    
    /**
     * execute close button command
     */
    private void do_btn_close()
    {
        hideConsole();        
    }
    
    
    /**
     * execute clear button command
     */
    private void do_btn_clear()
    {
        ta_messages.setText("");
    }
    
    
    /**
     * execute save button command
     */
    private void do_btn_save()
    {
        m_fileChooser.setSelectedFile( new File(m_messagesFilename) );
        int returnVal = m_fileChooser.showSaveDialog( this );
        if( returnVal==JFileChooser.APPROVE_OPTION )
        {
            File f = m_fileChooser.getSelectedFile();
            
            try
            {
                if( f.exists() )
                {
                    int res = JOptionPane.showConfirmDialog(
                                    this, 
                                    m_confimMessage,
                                    m_confimTitle,
                                    JOptionPane.YES_NO_OPTION, 
                                    JOptionPane.QUESTION_MESSAGE);
                    
                    if( res!=0 )
                        return;
                }
                
                IOHelper.saveTxtFile(f, ta_messages.getText(), false);
            }
            catch(Exception e)
            {
                System.err.println("Unable to save console text to file, error=" +
                    e.getMessage());
                e.printStackTrace();
            }            
        }
        else if( returnVal==JFileChooser.CANCEL_OPTION )
        {
            
        }
        else
        {
            
        }
    }
    
    
    /**
     * Does the actual check and flush of the textarea
     */
    private void do_it(boolean force)
    {
        String allText = ta_messages.getText();
        int len = allText.length();
            
        // Always scroll down to the last line
        ta_messages.setCaretPosition( len );
            
        // do we have a maximum characters limit?
        if( m_maxChars>0 )
        {
            // if we have exceeded the limit OR we are forced to flush the messages
            if( len > m_maxChars || force )
            {
                // should we save the messages before clearing them?
                if( m_autoSave )
                {
                    boolean append = true;
                    if( m_numFlushes==0 && !m_appendFirstTime )
                    {
                        append = false;
                    }
                        
                    try
                    {
                        IOHelper.saveTxtFile(m_traceFilename, allText, append);
                        m_numFlushes++;
                    }
                    catch(Exception e)
                    {
                        // not much to do if we have error here...
                        e.printStackTrace();            
                    }
                }
                    
                // crear the messages
                ta_messages.setText("");
            }
        }
    }
    
    
    // Variables declaration - do not modify//GEN-BEGIN:variables
    private javax.swing.JButton btn_clear;
    private javax.swing.JButton btn_close;
    private javax.swing.JButton btn_save;
    private javax.swing.JPanel button_panel;
    private javax.swing.JTextArea ta_messages;
    private javax.swing.JScrollPane text_scrollpane;
    // End of variables declaration//GEN-END:variables
    
    
    /**
     * PRIVATE CLASS
     * Filter to redirect the data to the textarea
     */
    private final class JTextAreaOutStream extends FilterOutputStream {

        
        public JTextAreaOutStream( OutputStream aStream )
        {
            super( aStream );
        }


        public void write( byte b[] )
            throws IOException
        {
            ta_messages.append( new String( b ) );
            do_it( false );
        }


        public void write( byte b[], int off, int len )
            throws IOException
        {
            ta_messages.append( new String( b , off , len ) );
            do_it( false );
        }

    };

}
