/*
 * JSplash.java - A simple, yet nice splash screen implementation for java
 *                Follows Sun recommendations for splash screen and logos:
 *                see "Designing Graphics for Corporate and Product Identity"
 *                http://java.sun.com/products/jlf/ed2/book/HIG.Graphics7.html
 *                Draws a black border of one pixel wide around the splash
 *                image. Also uses a simple progress bar that the user must
 *                "progress" manually in his code in order for it to work.
 *                Also, it has options for percent display or custom loading
 *                messages display.
 *
 * Copyright (c) 2004 Gregory Kotsaftis
 * gregkotsaftis@yahoo.com
 * http://zeus-jscl.sourceforge.net/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package gr.zeus.ui;

import gr.zeus.util.GUIUtils;
import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.MediaTracker;
import java.net.URL;
import javax.swing.BorderFactory;
import javax.swing.ImageIcon;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JProgressBar;
import javax.swing.JWindow;


public class JSplash extends JWindow {    

    private JProgressBar m_progress = null;
    private boolean m_progressBar = false;
    private boolean m_progressBarMessages = false;
    private boolean m_progressBarPercent = false;
    
    /**
     * USAGE:
     * url      = image for ImageIcon
     * progress = do we want a progress bar at all?
     * messages = if we want a progress bar, do we want to display messages
     *            inside the progress bar?
     * percent  = if we want a progress bar, do we want to display the percent
     *
     * NOTE: use only one, messages or percent
     */
    public JSplash(URL url, 
        boolean progress, boolean messages, boolean percent)
    {
        super();
        
        m_progressBar = progress;
        m_progressBarMessages = messages;
        m_progressBarPercent = percent;

        // build a panel with a black line for border, 
        // and set it as the content pane
        //
        JPanel panel = new JPanel();
        panel.setLayout( new BorderLayout() );
        panel.setBorder( BorderFactory.createLineBorder( Color.BLACK ) );        
        setContentPane( panel );
        
        // build a label and set it's icon
        //
        JLabel label = new JLabel();
        ImageIcon icon = new ImageIcon( url );        
        if( icon.getImageLoadStatus()!=MediaTracker.COMPLETE )
        {
            System.err.println("Cannot load splash screen: " + url);
            
            label.setText("Cannot load splash screen: " + url);
        }
        else
        {
            label.setIcon( icon );            
        }
        
        // build a progress bar
        //
        if( m_progressBar )
        {
            m_progress = new JProgressBar();

            if( m_progressBarMessages || m_progressBarPercent )
            {
                m_progress.setStringPainted( true );
            }
            else
            {
                m_progress.setStringPainted( false );
            }
            
            if( m_progressBarMessages && !m_progressBarPercent )
            {
                m_progress.setString( "" );                
            }
            
            m_progress.setMaximum( 100 );
            m_progress.setMinimum( 0 );
            m_progress.setValue( 0 );            
        }

        // add the components to the panel
        //
        getContentPane().add(label, BorderLayout.CENTER);
        
        if( m_progressBar )
        {
            getContentPane().add(m_progress, BorderLayout.SOUTH);
        }
        
        // validate, and display the components
        pack();
        
        // center on screen
        GUIUtils.centerOnScreen( this );
        /*
        Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
        Dimension splashSize = getContentPane().getPreferredSize();
        setLocation(screenSize.width / 2 - (splashSize.width / 2),
                    screenSize.height / 2 - (splashSize.height / 2));
        */
        
        // hide the panel for now...
        setVisible( false );
    }
    
    
    /**
     * Displays the splash screen
     */
    public void splashOn()
    {
        setVisible( true );
    }
    
    
    /**
     * Hides and disposes the splash screen
     */
    public void splashOff()
    {
        setVisible( false );
        dispose();
    }

    
    /**
     * Set the progress indicator (0...100)
     */
    public void setProgress(int value)
    {
        if( m_progressBar && value>=0 && value<=100 )
        {
            m_progress.setValue( value );
        }
    }

    
    /**
     * Set the progress indicator (0...100) and
     * a label to print inside the progress bar.
     */
    public void setProgress(int value, String msg)
    {
        setProgress( value );

        if( m_progressBarMessages && !m_progressBarPercent && msg!=null )
        {
            m_progress.setString( msg );
        }
    }
    
}
