/*
 * IOHelper.java - IO helper methods
 *
 * Copyright (c) 2004 Gregory Kotsaftis
 * gregkotsaftis@yahoo.com
 * http://zeus-jscl.sourceforge.net/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package gr.zeus.util;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;


public class IOHelper {
    
    
    /**
     * Create the complete directory structure
     * for a complete FOLDER pathname.
     */
    public static String createDirectoryTreeForFolder(String folder)
        throws IOException
    {
        if( folder==null )
        {
            throw new IOException("The requested folder is null!");
        }
        
        if( !folder.endsWith( File.separator ) )
        {
            folder += File.separator;
        }
        
        File f = new File( folder );
        if( !f.exists() )
        {
            boolean result = f.mkdirs(); // build all required directories!
            if( result==false )
            {
                throw new IOException("Cannot create folder structure for: " + folder);
            }
        }
        else if( !f.isDirectory() )
        {
            throw new IOException("Invalid directory/Cannot create directory: " + folder);
        }
        else
        {
            /* the folder already exists */
        }
        
        return( folder );
    }

    
    /**
     * Create the complete directory structure
     * for a complete FILE pathname.
     */
    public static void createDirectoryTreeForFile(String file)
        throws IOException
    {
        if( file==null )
        {
            throw new IOException("null file requested!");
        }
        
        String path = null;
        int sep = file.lastIndexOf( File.separatorChar );
        if( sep==-1 )
        {
            path = file;
        }
        else
        {
            path = file.substring(0, sep);
        }
        
        createDirectoryTreeForFolder( path );
    }
    
    
    /**
     * Determine if two filename paths refer to the same file
     *
     * See: http://javaalmanac.com/egs/java.io/Canonical.html
     */
    public static boolean isTheSameFile(String pathname1, String pathname2)
        throws IOException
    {
        File file1 = new File( pathname1 );
        File file2 = new File( pathname2 );
    
        // Normalize the paths
        file1 = file1.getCanonicalFile();
        file2 = file2.getCanonicalFile();
    
        return( file1.equals(file2) );
    }    
    
    
    /**
     * Determine if a file or directory exists
     */
    public static boolean fileOrDirectoryExists(String pathname)
    {
        File f = new File(pathname); 
        return( f.exists() );
    }
    

    /**
     * Determine if a file exists
     */
    public static boolean isFile(String pathname)
    {
        File f = new File( pathname );
        return( f.isFile() );
    }
    

    /**
     * Determine if a directory exists
     */
    public static boolean isDirectory(String pathname)
    {
        File f = new File( pathname );
        return( f.isDirectory() );
    }    
    

    /**
     * Save a string to a text file
     */
    public static void saveTxtFile(String pathname, String data, boolean append)
        throws IOException
    {
        saveTxtFile(new File(pathname), data, append);
    }
    
    
    /**
     * Save a string to a text file
     */
    public static void saveTxtFile(File f, String data, boolean append)
        throws IOException
    {
        BufferedWriter out = new BufferedWriter( new FileWriter(f, append) );
        out.write( data );
        out.close();
    }

    
    /**
     * Read a text file into a string
     */
    public static String readTxtFile(String pathname)
        throws IOException
    {
        return( readTxtFile( new File(pathname) ) );
    }
    
    
    /**
     * Read a text file into a string
     */
    public static String readTxtFile(File f)
        throws IOException
    {
        BufferedReader in = new BufferedReader( new FileReader(f) );
        String result = "";
        String str = null;
        while( (str = in.readLine()) != null )
        {
            result += str;
            result += "\n";
        }
        
        in.close();
        
        return( result );        
    }
    
    
    /**
     * Deletes all files and subdirectories under dir.
     * Returns true if all deletions were successful.
     * If a deletion fails, the method stops attempting
     * to delete and returns false.
     */
    public static boolean deleteDirectory(File dir)
    {
        if( dir.isDirectory() )
        {
            String[] children = dir.list();
            for(int i=0; i<children.length; i++)
            {
                boolean success = deleteDirectory( new File(dir, children[i]) );
                if( !success )
                {
                    return( false );
                }
            }
        }
    
        // This is a file or an empty directory, so just delete it
        return( dir.delete() );
    }
    
    
    /**
     * Deletes all files and subdirectories under dir.
     * Returns true if all deletions were successful.
     * If a deletion fails, the method stops attempting
     * to delete and returns false.
     */
    public static boolean deleteDirectory(String pathname)
    {
        return( deleteDirectory( new File(pathname) ) );        
    }

    
    /**
     * Deletes a file.
     * Returns true if successful.
     */
    public static boolean deleteFile(String pathname)
    {
        File f = new File( pathname );
        return( f.delete() );        
    }
    

    /**
     * Copies a file to another location/file
     * Returns true if file copied or 
     * false if not (possibly the file existed and 'overwrite' was not set).
     */
    public static boolean copyFile(String fromName, String toName, boolean overwrite)
        throws IOException
    {
        if( fromName==null )
            throw new IOException("source filename is null!");

        if( toName==null )
            throw new IOException("destination filename is null!");
        
        File fromFile = new File(fromName);	
        File toFile = new File(toName);
				
        return( copyFile(fromFile, toFile, overwrite) );
    }

	
    /**
     * Copies a file to another location/file
     * Returns true if file copied or 
     * false if not (possibly the file existed and 'overwrite' was not set).
     */
    public static boolean copyFile(File fromFile, File toFile, boolean overwrite)
        throws IOException
    {
        final int IO_BUF_SIZE = 1024 * 32;
    	
        if( fromFile==null )
            throw new IOException("source file is null!");
        
        if( toFile==null )
            throw new IOException("destination file is null!");
        
        // make sure that source file exists
        if( !fromFile.exists() )
            throw new IOException("no such source file: " + fromFile.getAbsoluteFile());
			
        if( !fromFile.isFile() )
            throw new IOException("can't copy directory: " + fromFile.getAbsoluteFile());
			
        if( !fromFile.canRead() )
            throw new IOException("source file is unreadable: " + fromFile.getAbsoluteFile());
			
        if( toFile.isDirectory() )
            toFile = new File(toFile, fromFile.getName());
			
        if( toFile.exists() )
        {
            if( !toFile.canWrite() )
                throw new IOException("destination file is unwriteable: " + toFile.getAbsoluteFile());
				
            // check if we should overwrite it
            if( !overwrite )
            {
            	  return( false );
            }
        }
        else
        {
            // if the file dosn't exist, check if the directory exists and is writeable
            // If getParent() returns null, then the directory is the current dir. so look
            // up the user.dir system property to find out what that is.
            String parent = toFile.getParent(); // the destination dir
            if( parent==null ) // if none use the current dir
                parent = System.getProperty("user.dir");
				
            File dir = new File( parent ); // convert it to a file
            if( !dir.exists() )
                throw new IOException("destination directory doesn't exist: " + parent);
				
            if( dir.isFile() )
                throw new IOException("destination is not a directory: " + parent);
				
            if( !dir.canWrite() )
                throw new IOException("destination directory is unwritable: " + parent);
        }
			
        // if we've gotten this far then everything is ok
        // so we copy the file one buffer of bytes at a time
        FileInputStream from = null; // Stream to read from source
        FileOutputStream to = null; // Stream to write to destination
        try
        {
            from = new FileInputStream(fromFile); // Create input stream
            to = new FileOutputStream(toFile); // Create output stream
            byte[] buffer = new byte[IO_BUF_SIZE]; // to hold file data
            int bytesRead;
			
            // read a chunk of bytes into the buffer then write them out
            // looping until we reach the EOF (when read() returns -1)
            // Note the combination of assignment and comparison in this 
            // while loop. This is a common I/O programming idiom.
            while( (bytesRead = from.read(buffer)) != -1 ) // read until EOF
            {
                to.write(buffer, 0, bytesRead); // write
            }
        }
        catch(IOException e)
        {
            throw( e );
        }
        finally
        {
            if( from!=null ) try { from.close(); } catch (IOException ignored) { }
            if( to!=null ) try { to.close(); } catch (IOException ignored) { }
        }
				
        return( true );
    }

}
